package meerkat.bulletinboard;

import com.google.common.util.concurrent.FutureCallback;
import com.google.common.util.concurrent.Futures;
import com.google.common.util.concurrent.ListeningScheduledExecutorService;
import com.google.common.util.concurrent.MoreExecutors;
import com.google.protobuf.ByteString;
import meerkat.bulletinboard.workers.singleserver.*;
import meerkat.comm.CommunicationException;
import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.protobuf.Voting.BulletinBoardClientParams;
import meerkat.util.BulletinBoardUtils;

import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * Created by Arbel Deutsch Peled on 28-Dec-15.
 *
 * This class implements the asynchronous Bulletin Board Client interface
 * It only handles a single Bulletin Board Server
 * If the list of servers contains more than one server: the server actually used is the first one
 * The class further implements a delayed access to the server after a communication error occurs
 */
public class SingleServerBulletinBoardClient extends SimpleBulletinBoardClient implements SubscriptionAsyncBulletinBoardClient {

    private final int MAX_RETRIES = 11;

    private ListeningScheduledExecutorService executorService;

    protected BatchDigest batchDigest;

    private long lastServerErrorTime;

    private final long failDelayInMilliseconds;

    private final long subscriptionIntervalInMilliseconds;

    /**
     * Notify the client that a job has failed
     * This makes new scheduled jobs be scheduled for a later time (after the given delay)
     */
    protected void fail() {

        // Update last fail time
        lastServerErrorTime = System.currentTimeMillis();

    }

    /**
     * This method adds a worker to the scheduled queue of the threadpool
     * If the server is in an accessible state: the job is submitted for immediate handling
     * If the server is not accessible: the job is scheduled for a later time
     * @param worker is the worker that should be scheduled for work
     * @param callback is the class containing callbacks for handling job completion/failure
     */
    protected void scheduleWorker(SingleServerWorker worker, FutureCallback callback){

        long timeSinceLastServerError = System.currentTimeMillis() - lastServerErrorTime;

        if (timeSinceLastServerError >= failDelayInMilliseconds) {

            // Schedule for immediate processing
            Futures.addCallback(executorService.submit(worker), callback);

        } else {

            // Schedule for processing immediately following delay expiry
            Futures.addCallback(executorService.schedule(
                    worker,
                    failDelayInMilliseconds - timeSinceLastServerError,
                    TimeUnit.MILLISECONDS),
                    callback);

        }

    }

    /**
     * Inner class for handling simple operation results and retrying if needed
     */
    class RetryCallback<T> implements FutureCallback<T> {

        private final SingleServerWorker worker;
        private final FutureCallback<T> futureCallback;

        public RetryCallback(SingleServerWorker worker, FutureCallback<T> futureCallback) {
            this.worker = worker;
            this.futureCallback = futureCallback;
        }

        @Override
        public void onSuccess(T result) {
            futureCallback.onSuccess(result);
        }

        @Override
        public void onFailure(Throwable t) {

            // Notify client about failure
            fail();

            // Check if another attempt should be made

            worker.decMaxRetry();

            if (worker.isRetry()) {
                // Perform another attempt
                scheduleWorker(worker, this);
            } else {
                // No more retries: notify caller about failure
                futureCallback.onFailure(t);
            }

        }

    }

    /**
     * This callback ties together all the per-batch-data callbacks into a single callback
     * It reports success back to the user only if all of the batch-data were successfully posted
     * If any batch-data fails to post: this callback reports failure
     */
    class PostBatchDataListCallback implements FutureCallback<Boolean> {

        private final FutureCallback<Boolean> callback;

        private AtomicInteger batchDataRemaining;
        private AtomicBoolean aggregatedResult;

        public PostBatchDataListCallback(int batchDataLength, FutureCallback<Boolean> callback) {

            this.callback = callback;
            this.batchDataRemaining = new AtomicInteger(batchDataLength);
            this.aggregatedResult = new AtomicBoolean(false);

        }

        @Override
        public void onSuccess(Boolean result) {

            if (result){
                this.aggregatedResult.set(true);
            }

            if (batchDataRemaining.decrementAndGet() == 0){
                callback.onSuccess(this.aggregatedResult.get());
            }
        }

        @Override
        public void onFailure(Throwable t) {

            // Notify caller about failure
            callback.onFailure(t);

        }
    }

    /**
     * This callback ties together the different parts of a CompleteBatch as they arrive from the server
     * It assembles a CompleteBatch from the parts and sends it to the user if all parts arrived
     * If any part fails to arrive: it invokes the onFailure method
     */
    class CompleteBatchReadCallback {

        private final FutureCallback<CompleteBatch> callback;

        private List<BatchData> batchDataList;
        private BulletinBoardMessage batchMessage;

        private AtomicInteger remainingQueries;
        private AtomicBoolean failed;

        public CompleteBatchReadCallback(FutureCallback<CompleteBatch> callback) {

            this.callback = callback;

            remainingQueries = new AtomicInteger(2);
            failed = new AtomicBoolean(false);

        }

        protected void combineAndReturn() {

            final String[] prefixes = {
                    BulletinBoardConstants.BATCH_ID_TAG_PREFIX,
                    BulletinBoardConstants.BATCH_TAG};

            if (remainingQueries.decrementAndGet() == 0){

                String batchIdStr = BulletinBoardUtils.findTagWithPrefix(batchMessage, BulletinBoardConstants.BATCH_ID_TAG_PREFIX);

                if (batchIdStr == null){
                    callback.onFailure(new CommunicationException("Server returned invalid message with no Batch ID tag"));
                }

                BeginBatchMessage beginBatchMessage =
                        BeginBatchMessage.newBuilder()
                                .setSignerId(batchMessage.getSig(0).getSignerId())
                                .setBatchId(Integer.parseInt(batchIdStr))
                                .addAllTag(BulletinBoardUtils.removePrefixTags(batchMessage, Arrays.asList(prefixes)))
                                .build();
                callback.onSuccess(new CompleteBatch(beginBatchMessage, batchDataList, batchMessage.getSig(0)));
            }

        }

        protected void fail(Throwable t) {
            if (failed.compareAndSet(false, true)) {
                callback.onFailure(t);
            }
        }

        /**
         * @return a FutureCallback for the Batch Data List that ties to this object
         */
        public FutureCallback<List<BatchData>> asBatchDataListFutureCallback() {
            return new FutureCallback<List<BatchData>>() {

                @Override
                public void onSuccess(List<BatchData> result) {
                    batchDataList = result;

                    combineAndReturn();
                }

                @Override
                public void onFailure(Throwable t) {
                    fail(t);
                }

            };
        }

        /**
         * @return a FutureCallback for the Bulletin Board Message that ties to this object
         */
        public FutureCallback<List<BulletinBoardMessage>> asBulletinBoardMessageListFutureCallback() {
            return new FutureCallback<List<BulletinBoardMessage>>() {

                @Override
                public void onSuccess(List<BulletinBoardMessage> result) {
                    if (result.size() < 1){
                        onFailure(new IllegalArgumentException("Server returned empty message list"));
                        return;
                    }

                    batchMessage = result.get(0);

                    combineAndReturn();
                }

                @Override
                public void onFailure(Throwable t) {
                    fail(t);
                }
            };
        }

    }


    /**
     * Inner class for handling returned values of subscription operations
     * This class's methods also ensure continued operation of the subscription
     */
    class SubscriptionCallback implements FutureCallback<List<BulletinBoardMessage>> {

        private SingleServerReadMessagesWorker worker;
        private final FutureCallback<List<BulletinBoardMessage>> callback;

        private MessageFilterList.Builder filterBuilder;

        public SubscriptionCallback(SingleServerReadMessagesWorker worker, FutureCallback<List<BulletinBoardMessage>> callback) {
            this.worker = worker;
            this.callback = callback;
            filterBuilder = worker.getPayload().toBuilder();

        }

        @Override
        public void onSuccess(List<BulletinBoardMessage> result) {

            // Report new messages to user
            callback.onSuccess(result);

            // Remove last filter from list (MIN_ENTRY one)
            filterBuilder.removeFilter(filterBuilder.getFilterCount() - 1);

            // Add updated MIN_ENTRY filter (entry number is successor of last received entry's number)
            filterBuilder.addFilter(MessageFilter.newBuilder()
                    .setType(FilterType.MIN_ENTRY)
                    .setEntry(result.get(result.size() - 1).getEntryNum() + 1)
                    .build());

            // Create new worker with updated task
            worker = new SingleServerReadMessagesWorker(worker.serverAddress, filterBuilder.build(), 1);

            // Schedule the worker
            scheduleWorker(worker, this);

        }

        @Override
        public void onFailure(Throwable t) {

            // Notify client about failure
            fail();

            // Notify caller about failure and terminate subscription
            callback.onFailure(t);
        }
    }

    public SingleServerBulletinBoardClient(ListeningScheduledExecutorService executorService,
                                           long failDelayInMilliseconds,
                                           long subscriptionIntervalInMilliseconds) {

        this.executorService = executorService;

        this.failDelayInMilliseconds = failDelayInMilliseconds;
        this.subscriptionIntervalInMilliseconds = subscriptionIntervalInMilliseconds;

        // Set server error time to a time sufficiently in the past to make new jobs go through
        lastServerErrorTime = System.currentTimeMillis() - failDelayInMilliseconds;

    }

    public SingleServerBulletinBoardClient(int threadPoolSize, long failDelayInMilliseconds, long subscriptionIntervalInMilliseconds) {

        this(MoreExecutors.listeningDecorator(Executors.newScheduledThreadPool(threadPoolSize)),
                failDelayInMilliseconds,
                subscriptionIntervalInMilliseconds);

    }

    /**
     * Stores database location, initializes the web Client and
     * @param clientParams contains the data needed to access the DBs
     */
    @Override
    public void init(BulletinBoardClientParams clientParams) {

        // Perform usual setup
        super.init(clientParams);

        // Wrap the Digest into a BatchDigest
        batchDigest = new GenericBatchDigest(digest);

        // Remove all but first DB address
        String dbAddress = meerkatDBs.get(0);
        meerkatDBs = new LinkedList<>();
        meerkatDBs.add(dbAddress);

    }

    @Override
    public MessageID postMessage(BulletinBoardMessage msg, FutureCallback<Boolean> callback) {

        // Create worker with redundancy 1 and MAX_RETRIES retries
        SingleServerPostMessageWorker worker = new SingleServerPostMessageWorker(meerkatDBs.get(0), msg, MAX_RETRIES);

        // Submit worker and create callback
        scheduleWorker(worker, new RetryCallback<>(worker, callback));

        // Calculate the correct message ID and return it
        batchDigest.reset();
        batchDigest.update(msg.getMsg());
        return batchDigest.digestAsMessageID();

    }

    private class PostBatchDataCallback implements FutureCallback<Boolean> {

        private final CompleteBatch completeBatch;
        private final FutureCallback<Boolean> callback;

        public PostBatchDataCallback(CompleteBatch completeBatch, FutureCallback<Boolean> callback) {
            this.completeBatch = completeBatch;
            this.callback = callback;
        }

        @Override
        public void onSuccess(Boolean msg) {
            closeBatch(
                    completeBatch.getCloseBatchMessage(),
                    callback
            );
        }

        @Override
        public void onFailure(Throwable t) {
            callback.onFailure(t);
        }

    }

    private class BeginBatchCallback implements FutureCallback<Boolean> {

        private final CompleteBatch completeBatch;
        private final FutureCallback<Boolean> callback;

        public BeginBatchCallback(CompleteBatch completeBatch, FutureCallback<Boolean> callback) {
            this.completeBatch = completeBatch;
            this.callback = callback;
        }

        @Override
        public void onSuccess(Boolean msg) {

            postBatchData(
                    completeBatch.getBeginBatchMessage().getSignerId(),
                    completeBatch.getBeginBatchMessage().getBatchId(),
                    completeBatch.getBatchDataList(),
                    0,
                    new PostBatchDataCallback(completeBatch,callback));
        }

        @Override
        public void onFailure(Throwable t) {
            callback.onFailure(t);
        }
    }

    @Override
    public MessageID postBatch(CompleteBatch completeBatch, FutureCallback<Boolean> callback) {

        beginBatch(
                completeBatch.getBeginBatchMessage(),
                new BeginBatchCallback(completeBatch, callback)
        );

        batchDigest.update(completeBatch);

        return batchDigest.digestAsMessageID();

    }

    @Override
    public void beginBatch(BeginBatchMessage beginBatchMessage, FutureCallback<Boolean> callback) {

        // Create worker with redundancy 1 and MAX_RETRIES retries
        SingleServerBeginBatchWorker worker =
                new SingleServerBeginBatchWorker(meerkatDBs.get(0), beginBatchMessage, MAX_RETRIES);

        // Submit worker and create callback
        scheduleWorker(worker, new RetryCallback<>(worker, callback));

    }

    @Override
    public void postBatchData(ByteString signerId, int batchId, List<BatchData> batchDataList,
                          int startPosition, FutureCallback<Boolean> callback) {

        BatchMessage.Builder builder = BatchMessage.newBuilder()
                .setSignerId(signerId)
                .setBatchId(batchId);

        // Create a unified callback to aggregate successful posts

        PostBatchDataListCallback listCallback = new PostBatchDataListCallback(batchDataList.size(), callback);

        // Iterate through data list

        for (BatchData data : batchDataList) {
            builder.setSerialNum(startPosition).setData(data);

            // Create worker with redundancy 1 and MAX_RETRIES retries
            SingleServerPostBatchWorker worker =
                    new SingleServerPostBatchWorker(meerkatDBs.get(0), builder.build(), MAX_RETRIES);

            // Create worker with redundancy 1 and MAX_RETRIES retries
            scheduleWorker(worker, new RetryCallback<>(worker, listCallback));

            // Increment position in batch
            startPosition++;
        }

    }

    @Override
    public void postBatchData(ByteString signerId, int batchId, List<BatchData> batchDataList, FutureCallback<Boolean> callback) {

        postBatchData(signerId, batchId, batchDataList, 0, callback);

    }

    @Override
    public void postBatchData(byte[] signerId, int batchId, List<BatchData> batchDataList,
                              int startPosition, FutureCallback<Boolean> callback) {

        postBatchData(ByteString.copyFrom(signerId), batchId, batchDataList, startPosition, callback);

    }

    @Override
    public void postBatchData(byte[] signerId, int batchId, List<BatchData> batchDataList, FutureCallback<Boolean> callback) {

        postBatchData(signerId, batchId, batchDataList, 0, callback);

    }

    @Override
    public void closeBatch(CloseBatchMessage closeBatchMessage, FutureCallback<Boolean> callback) {

        // Create worker with redundancy 1 and MAX_RETRIES retries
        SingleServerCloseBatchWorker worker =
                new SingleServerCloseBatchWorker(meerkatDBs.get(0), closeBatchMessage, MAX_RETRIES);

        // Submit worker and create callback
        scheduleWorker(worker, new RetryCallback<>(worker, callback));

    }

    @Override
    public void getRedundancy(MessageID id, FutureCallback<Float> callback) {

        // Create worker with no retries
        SingleServerGetRedundancyWorker worker = new SingleServerGetRedundancyWorker(meerkatDBs.get(0), id, 1);

        // Submit job and create callback
        scheduleWorker(worker, new RetryCallback<>(worker, callback));

    }

    @Override
    public void readMessages(MessageFilterList filterList, FutureCallback<List<BulletinBoardMessage>> callback) {

        // Create job with no retries
        SingleServerReadMessagesWorker worker = new SingleServerReadMessagesWorker(meerkatDBs.get(0), filterList, 1);

        // Submit job and create callback
        scheduleWorker(worker, new RetryCallback<>(worker, callback));

    }

    @Override
    public void readBatch(BatchSpecificationMessage batchSpecificationMessage, FutureCallback<CompleteBatch> callback) {

        // Create job with no retries for retrieval of the Bulletin Board Message that defines the batch

        MessageFilterList filterList = MessageFilterList.newBuilder()
                .addFilter(MessageFilter.newBuilder()
                        .setType(FilterType.TAG)
                        .setTag(BulletinBoardConstants.BATCH_TAG)
                        .build())
                .addFilter(MessageFilter.newBuilder()
                        .setType(FilterType.TAG)
                        .setTag(BulletinBoardConstants.BATCH_ID_TAG_PREFIX + batchSpecificationMessage.getBatchId())
                        .build())
                .addFilter(MessageFilter.newBuilder()
                        .setType(FilterType.SIGNER_ID)
                        .setId(batchSpecificationMessage.getSignerId())
                        .build())
                .build();

        SingleServerReadMessagesWorker messageWorker = new SingleServerReadMessagesWorker(meerkatDBs.get(0), filterList, 1);

        // Create job with no retries for retrieval of the Batch Data List
        SingleServerReadBatchWorker batchWorker = new SingleServerReadBatchWorker(meerkatDBs.get(0), batchSpecificationMessage, 1);

        // Create callback that will combine the two worker products
        CompleteBatchReadCallback completeBatchReadCallback = new CompleteBatchReadCallback(callback);

        // Submit jobs with wrapped callbacks
        scheduleWorker(messageWorker, new RetryCallback<>(messageWorker, completeBatchReadCallback.asBulletinBoardMessageListFutureCallback()));
        scheduleWorker(batchWorker, new RetryCallback<>(batchWorker, completeBatchReadCallback.asBatchDataListFutureCallback()));

    }

    @Override
    public void querySync(SyncQuery syncQuery, FutureCallback<SyncQueryResponse> callback) {

        SingleServerQuerySyncWorker worker = new SingleServerQuerySyncWorker(meerkatDBs.get(0), syncQuery, MAX_RETRIES);

        scheduleWorker(worker, new RetryCallback<>(worker, callback));

    }

    @Override
    public void subscribe(MessageFilterList filterList, long startEntry, FutureCallback<List<BulletinBoardMessage>> callback) {
        // Remove all existing MIN_ENTRY filters and create new one that starts at 0

        MessageFilterList.Builder filterListBuilder = filterList.toBuilder();

        Iterator<MessageFilter> iterator = filterListBuilder.getFilterList().iterator();
        while (iterator.hasNext()) {
            MessageFilter filter = iterator.next();

            if (filter.getType() == FilterType.MIN_ENTRY){
                iterator.remove();
            }
        }
        filterListBuilder.addFilter(MessageFilter.newBuilder()
                .setType(FilterType.MIN_ENTRY)
                .setEntry(startEntry)
                .build());

        // Create job with no retries
        SingleServerReadMessagesWorker worker = new SingleServerReadMessagesWorker(meerkatDBs.get(0), filterListBuilder.build(), MAX_RETRIES);

        // Submit job and create callback that retries on failure and handles repeated subscription
        scheduleWorker(worker, new RetryCallback<>(worker, new SubscriptionCallback(worker, callback)));
    }

    @Override
    public void subscribe(MessageFilterList filterList, FutureCallback<List<BulletinBoardMessage>> callback) {
        subscribe(filterList, 0, callback);
    }

    @Override
    public void close() {

        super.close();

        executorService.shutdown();

    }
}
