package meerkat.voting.output;

import com.google.common.util.concurrent.FutureCallback;
import meerkat.protobuf.Voting.BallotSecrets;
import meerkat.protobuf.Voting.PlaintextBallot;
import meerkat.protobuf.Voting.SignedEncryptedBallot;
import meerkat.voting.controller.callbacks.ControllerCallback;
import meerkat.voting.output.outputcommands.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.concurrent.LinkedBlockingQueue;

/**
 * Created by hai on 27/06/16.
 */
public abstract class AsyncRunnableOutputDevice implements BallotOutputDevice, Runnable  {

    private Logger logger;
    private LinkedBlockingQueue<OutputCommand> queue;
    private volatile boolean shutDownHasBeenCalled;

    public AsyncRunnableOutputDevice() {
        logger = LoggerFactory.getLogger(AsyncRunnableOutputDevice.class);
        logger.info("AsyncRunnableOutputDevice is constructed");
        queue = new LinkedBlockingQueue<>();
        shutDownHasBeenCalled = false;
    }

    @Override
    public void run () {
        logger.info("starts running");
        while (! wasShutDownCalled()) {
            try {
                OutputCommand command = queue.take();
                handleSingleCommand(command);
            }
            catch (InterruptedException e) {
                logger.warn ("Interrupted while reading from command queue " + e);
            }
        }
    }

    private boolean wasShutDownCalled () {
        return shutDownHasBeenCalled;
    }

    @Override
    public void callShutDown() {
        logger.info("callShutDown command has been called");
        shutDownHasBeenCalled = true;
        queue.clear();
    }


    abstract void doCommitToBallot(CommitOutputCommand command);

    abstract void doAudit(AuditOutputCommand command);

    abstract void doCastBallot(CastOutputCommand command);

    abstract void doCancel(CancelOutputCommand command);


    private void handleSingleCommand(OutputCommand command) {
        if (command instanceof CommitOutputCommand) {
            doCommitToBallot((CommitOutputCommand)command);
        }
        else if (command instanceof AuditOutputCommand) {
            doAudit((AuditOutputCommand)command);
        }
        else if (command instanceof CastOutputCommand) {
            doCastBallot((CastOutputCommand)command);
        }
        else if (command instanceof CancelOutputCommand) {
            doCancel((CancelOutputCommand)command);
        }
        else {
            String errorMessage = "handleSingleCommand: unknown type of OutputCommand received: " +
                    command.getClass().getName();
            logger.error(errorMessage);
            throw new RuntimeException(errorMessage);
        }
    }


    @Override
    public void commitToBallot(PlaintextBallot plaintextBallot,
                               SignedEncryptedBallot signedEncryptedBallot,
                               FutureCallback<Void> callback) {
        logger.debug("Output interface call to commit to ballot");
        queue.add(new CommitOutputCommand(plaintextBallot, signedEncryptedBallot, (ControllerCallback)callback));
    }

    @Override
    public void audit(BallotSecrets ballotSecrets, FutureCallback<Void> callback) {
        logger.debug("an interface call to audit");
        queue.add(new AuditOutputCommand(ballotSecrets, (ControllerCallback)callback));
    }

    @Override
    public void castBallot(FutureCallback<Void> callback) {
        logger.debug("an interface call to cast ballot");
        queue.add(new CastOutputCommand((ControllerCallback)callback));
    }

    @Override
    public void cancelBallot(FutureCallback<Void> callback) {
        logger.debug("an interface call to cancel the output");
        queue.add(new CancelOutputCommand((ControllerCallback)callback));
    }


}
