package meerkat.voting.ui;

import com.google.common.util.concurrent.FutureCallback;
import com.google.protobuf.ByteString;
import meerkat.protobuf.Voting.*;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.*;
import java.util.concurrent.LinkedBlockingQueue;

import meerkat.voting.controller.callbacks.*;
import meerkat.voting.ui.uicommands.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import static java.lang.System.in;


public class SystemConsoleUI implements VotingBoothUI, Runnable {

    private BufferedReader bufferedReader;
    private LinkedBlockingQueue<UICommand> queue;
    private final Logger logger;
    private final int tickDurationInMillisec = 10;
    private Date startWaitingTime;

    private volatile boolean shutDownHasBeenCalled;



    public SystemConsoleUI() {
        logger = LoggerFactory.getLogger(SystemConsoleUI.class);
        logger.info("A VB UI console is constructed");
        queue = new LinkedBlockingQueue<>();
        bufferedReader = new BufferedReader(new InputStreamReader(in));

        startWaitingTime = null;
        Timer timer = new Timer();
        timer.scheduleAtFixedRate(new TickerTimerTask(queue), new Date(), tickDurationInMillisec);

        shutDownHasBeenCalled = false;
    }


    @Override
    public void run () {
        logger.info("UI starts running");
        while (! wasShutDownCalled()) {
            try {
                UICommand command = queue.take();
                handleSingleCommand(command);
            }
            catch (InterruptedException e) {
                logger.warn ("Interrupted while reading from command queue " + e);
            }
        }
    }


    @Override
    public void callShutDown() {
        logger.info("callShutDown command has been called");
        shutDownHasBeenCalled = true;
        queue.clear();
    }

    private void handleSingleCommand(UICommand command) {
        if (!(command instanceof TickCommand)) {
            if (startWaitingTime != null) {
                stopWaiting();
            }
        }

        if (command instanceof StartSessionUICommand) {
            doShowWelcomeScreen((StartSessionUICommand)command);
        }
        else if (command instanceof ChannelChoiceUICommand) {
            doAskChannelChoiceQuestions((ChannelChoiceUICommand)command);
        }
        else if (command instanceof RaceVotingUICommand) {
            doAskVotingQuestions((RaceVotingUICommand)command);
        }
        else if (command instanceof CastOrAuditUICommand) {
            doCastOrAudit ((CastOrAuditUICommand)command);
        }
        else if (command instanceof FatalErrorUICommand) {
            doFatalError((FatalErrorUICommand)command);
        }
        else if (command instanceof WaitForFinishUICommand) {
            doWaitForFinish((WaitForFinishUICommand)command);
        }
        else if (command instanceof TickCommand) {
            doTick ();
        }
        else {
            String errorMessage = "handleSingleCommand: unknown type of UICommand received: " +
                    command.getClass().getName();
            logger.error(errorMessage);
            throw new RuntimeException(errorMessage);
        }
    }


    @Override
    public void startNewVoterSession(FutureCallback<Void> callback) {
        logger.debug("UI interface call to startNewVoterSession");
        queue.add(new StartSessionUICommand((ControllerCallback)callback));
    }

    private void doShowWelcomeScreen(StartSessionUICommand command) {
        logger.debug("UI entered doShowWelcomeScreen");
        System.out.println("Welcome, new voter!");
        waitForEnter(null);
        ControllerCallback callback = command.getCallback();
        assert (callback instanceof NewVoterCallback);
        callback.onSuccess(null);
    }

    private void stopWaiting () {
        System.out.println ();
        startWaitingTime = null;

    }
    private void waitForEnter(String message) {
        if (message != null) {
            System.out.println(message);
        }
        System.out.println("\nPress ENTER to proceed.\n");
        String t = null;
        while (t == null) {
            try {
                t = readInputLine();
            }
            catch (IOException e) {
                String errorMessage = "waitForEnter: threw an IOException: " + e;
                logger.error(errorMessage);
                System.err.println(errorMessage);
            }
        }
    }

    @Override
    public void chooseChannel(List<BallotQuestion> questions, FutureCallback<List<BallotAnswer>> callback) {
        logger.debug("UI interface call to chooseChannel");
        ChannelChoiceUICommand command = new ChannelChoiceUICommand(questions, (ControllerCallback)callback);
        queue.add(command);
    }

    private void doAskChannelChoiceQuestions (ChannelChoiceUICommand command) {
        logger.debug("UI: doAskChannelChoiceQuestions");
        System.out.println("Showing questions for choosing channel:\n");
        try {
            List<BallotAnswer> answers = askVoterForAnswers(command.getQuestions());
            command.getCallback().onSuccess(answers);
        }
        catch (VoterCancelThrowable e) {
            command.getCallback().onFailure(e);
        }
        catch (IOException e) {
            String errorMessage = "Channel choice failed due to IOException: " + e;
            logger.error (errorMessage);
            System.err.println(errorMessage);
            command.getCallback().onFailure(e);
        }
    }

    @Override
    public void askVoterQuestions(List<BallotQuestion> questions, FutureCallback<List<BallotAnswer>> callback) {
        logger.debug("UI interface call to chooseChannel");
        queue.add(new RaceVotingUICommand(questions, (ControllerCallback)callback));
    }

    private void doAskVotingQuestions (RaceVotingUICommand command) {
        logger.debug("UI: doAskVotingQuestions");
        System.out.println("Showing questions for race voting:\n");
        try {
            List<BallotAnswer> answers = askVoterForAnswers(command.getQuestions());
            command.getCallback().onSuccess(answers);
        }
        catch (VoterCancelThrowable e) {
            command.getCallback().onFailure(e);
        }
        catch (IOException e) {
            String errorMessage = "Asking voting questions failed due to IOException: " + e;
            logger.error (errorMessage);
            System.err.println(errorMessage);
            command.getCallback().onFailure(e);
        }
    }

    @Override
    public void castOrAudit(FutureCallback<FinalizeBallotChoice> callback) {
        logger.debug("UI interface call to castOrAudit");
        queue.add(new CastOrAuditUICommand((ControllerCallback)callback));
    }

    private void doCastOrAudit(CastOrAuditUICommand command) {
        logger.debug("UI entered doCastOrAudit");
        System.out.println ("Finalizing your vote. Do you wish to (C)ast or (A)udit?");

        FinalizeBallotChoice fChoice;

        try {
            String s = readInputLine();
            if (s.equals("cast") || s.equals("c")) {
                fChoice = FinalizeBallotChoice.CAST;
            }
            else if (s.equals("audit") || s.equals("a")) {
                fChoice = FinalizeBallotChoice.AUDIT;
            }
            else {
                throw new IllegalArgumentException("UI could not understand the answer for cast/audit question '" + s + "'");
            }
            ControllerCallback callback = command.getCallback();
            assert (callback instanceof CastOrAuditCallback);
            ((CastOrAuditCallback)callback).onSuccess(fChoice);
        }
        catch (IllegalArgumentException|IOException e) {
            String errorMessage = "doCastOrAudit: some error with reading input from console. details: " + e;
            logger.error(errorMessage);
            command.getCallback().onFailure(e);
        }
    }


    @Override
    public void notifyVoterToWaitForFinish(UIElement message, FutureCallback<Void> callback) {
        logger.debug("UI interface call to notifyVoterToWaitForFinish");
        queue.add(new WaitForFinishUICommand(message, (ControllerCallback)callback));
    }

    public void doWaitForFinish (WaitForFinishUICommand command) {
        logger.debug("UI entered doWaitForFinish");

        startWaitingTime = new Date();

        UIElement message = command.getMessage();
        String messageString;
        if (message.getType() != UIElementDataType.TEXT) {
            messageString = "Default message: encountered an error. System halting";
        } else {
            messageString = bytesToString(message.getData());
        }
        System.out.println(messageString);
        System.out.print ("Waiting : .");
    }

    @Override
    public void showErrorMessageAndHalt(UIElement errorMessage, FutureCallback<Boolean> callback) {
        logger.debug("UI interface call to showErrorMessageAndHalt");
        throw new UnsupportedOperationException("Not implemented becuase currently not sure if we ever use it.");
    }


    @Override
    public void showErrorMessageWithButtons(UIElement errorMessage, UIElement[] buttonLabels, FutureCallback<Integer> callback) {
        logger.debug("UI interface call to showErrorMessageWithButtons");
        queue.clear();
        queue.add(new FatalErrorUICommand(errorMessage, buttonLabels, (ControllerCallback)callback));
    }

    private void doFatalError (FatalErrorUICommand command) {
        logger.debug("UI entered doFatalError");

        UIElement errorMessage = command.getErrorMessage();
        String errorMessageString;
        if (errorMessage.getType() != UIElementDataType.TEXT) {
            errorMessageString = "Default message: encountered an error. System halting";
        } else {
            errorMessageString = bytesToString(errorMessage.getData());
        }

        UIElement[] buttonLabels = command.getButtonLabels();
        String[] buttonLabelStrings = new String[buttonLabels.length];
        for (int i = 0; i < buttonLabels.length; ++i) {
            if (buttonLabels[i].getType() != UIElementDataType.TEXT) {
                buttonLabelStrings[i] = "<NON_TEXT>";
            } else {
                buttonLabelStrings[i] = bytesToString(errorMessage.getData());
            }
        }

        System.out.println(errorMessageString);
        for (int i = 0; i < buttonLabelStrings.length; ++i) {
            System.out.println("" + i + " - " + buttonLabelStrings[i]);
        }

        try {
            String s = readInputLine();
            Integer chosenButton = new Integer(s);
            command.getCallback().onSuccess(chosenButton);
        }
        catch (IOException e) {
            String err = "doFatalError: some error with reading input from console. details: " + e;
            logger.error(err);
            command.getCallback().onFailure(e);
        }

    }


    private void doTick () {
        if (startWaitingTime != null) {
            System.out.print (".");  // still waiting
        }
    }

    private String readInputLine() throws IOException{
        String s;
        try {
            s = this.bufferedReader.readLine();
            if (null == s) {
                throw new IOException();
            }
        } catch (IOException e) {
            String errorMessage = "readInputLine: some error with reading input from console. details: " + e;
            logger.error(errorMessage);
            throw new IOException(e);
        }
        return s;
    }


    private void assertQuestionsAreValid (List<BallotQuestion> questions) {
        for (int index = 0; index < questions.size(); ++index) {
            BallotQuestion question = questions.get(index);
            if (question.getIsMandatory()) {
                String errorMessage = "askVoterQuestions: question number " + index + " is marked as mandatory";
                logger.error(errorMessage);
                throw new UnsupportedOperationException(errorMessage);
            }
            if (!isQuestionOnlyText(question)) {
                String errorMessage = "askVoterQuestions: question number " + index + " is not only text";
                logger.error(errorMessage);
                throw new UnsupportedOperationException(errorMessage);
            }
        }
    }

    private List<BallotAnswer> askVoterForAnswers(List<BallotQuestion> questions) throws VoterCancelThrowable, IOException {

        assertQuestionsAreValid (questions);

        List<BallotAnswer> answers = new ArrayList<>();
        int index = 0;
        while (index < questions.size()) {
            BallotQuestion question = questions.get(index);
            System.out.println("Question number " + index);
            showQuestionOnScreen(question);

            System.out.println("UI screen: Enter your answer. You can also type '(b)ack' or '(c)ancel' or '(s)kip");
            String s = readInputLine();

            if ((s.equals("cancel") || s.equals("c")) || (index == 0 && (s.equals("back") || s.equals("b")))) {
                throw new VoterCancelThrowable();
            }
            else if (s.equals("back") || s.equals("b")) {
                --index;
                answers.remove(index);
            }
            else if (s.equals("skip") || s.equals("s")) {
                answers.add(translateStringAnswerToProtoBufMessageAnswer(""));
                ++index;
            }
            else {
                answers.add(translateStringAnswerToProtoBufMessageAnswer(s));
                ++index;
            }
        }
        return answers;
    }

    //show question on the screen for the voter
    private void showQuestionOnScreen(BallotQuestion question) {
        if (!isQuestionOnlyText(question)) {
            System.err.println("debug: an element in question is not of TEXT type");
            throw new UnsupportedOperationException();
        }

        System.out.println("Question text: " + bytesToString(question.getQuestion().getData()));
        System.out.println("Description: " + bytesToString(question.getDescription().getData()));
        int answerIndex = 0;
        for (UIElement answer : question.getAnswerList()) {
            ++answerIndex;
            System.out.println("Answer " + answerIndex + ": " + bytesToString(answer.getData()));
        }
    }

    private boolean isQuestionOnlyText (BallotQuestion question) {
        boolean isText = true;
        if (question.getQuestion().getType() != UIElementDataType.TEXT
                || question.getDescription().getType() != UIElementDataType.TEXT) {
            isText = false;
        }
        for (UIElement answer : question.getAnswerList()) {
            if (answer.getType() != UIElementDataType.TEXT) {
                isText = false;
            }
        }
        return isText;
    }


    private BallotAnswer translateStringAnswerToProtoBufMessageAnswer(String s) {
        BallotAnswer.Builder bab = BallotAnswer.newBuilder();
        StringTokenizer st = new StringTokenizer(s);
        while (st.hasMoreTokens()) {
            bab.addAnswer(Integer.parseInt(st.nextToken()));
        }
        return bab.build();
    }

    /*
     * Returns the UTF8 decoding of byte-string data
     */
    private static String bytesToString(ByteString data) {
        return data.toStringUtf8();
    }


    private boolean wasShutDownCalled () {
        return shutDownHasBeenCalled;
    }


}
